/**
Modified by Mock Suwannatat
mock@cs.ucsb.edu
for MAT 259 Project 1
*/

/* SPL_xmlParse 1.11.2008 - Charlie Roberts
**
** This code will download an UTF8 xml file and display the xml elements and their values.
** It also converts text nodes to integer values for the deweyClass children. Once converted
** to integers, this data can then be used to position rectangles, circles, or any other type of
** visual element.
**
** Processing can't handle printing out all the nodes to the console window (it crashes on my computer)
** so I added a couple of lines of code that tells Processing to only print the first 50 elements
** if there are more than 50 transactions in the xml file. When I convert the data to rectangles instead
** of printing the text, Processing has no problem with it. We'll look at that next week.
*/  


import proxml.*;      // imports the proxml library
import java.util.*;

int w = 1000;
int h = 500;
int topMargin = 40;
int leftMargin = 100;
int rightMargin = leftMargin;
int lineHeight = 20;
int textGap = 10;
int textYOffset = 16;
boolean parsingDone = false;
int barHue = 68; //124; //25;

//String XML_location = "http://www.charlie-roberts.com/259/utfConvert_16to8.php?&year=2008&month=01&day=02&hour=14&inOut=out";
String XML_location = "xml3.xml";
XMLInOut xml;         // creates a variable named 'xml' that will hold an XMLInOut object
PFont font, fontBig; 
HashMap wordMap;
int wordCount = 0;
int maxCount = 0; 
final int TOP_N = 20;
int count = 0;
WordStat[] popWords; 
float alignNoon = -1, alignEvening = -1;
boolean movingRight = false, movingLeft = false;
boolean singleBarMode = true;
int stepNo = 0;

void setup() {
  size(w, h);
  xml = new XMLInOut(this);                                                              // creates an XMLInOut object and stores it in the 'xml' variable
  xml.loadElement(XML_location);    // loads the xml file into the XMLInOut object stored in the variable 'xml' 
  font = loadFont("ArialMT-12.vlw");
  fontBig = loadFont("HelveticaNeue-CondensedBold-16.vlw"); 
  
  textFont(fontBig, 16);
  fill(#880000);
  background(#ffffff);
  text("Loading. Please wait ...", 50, 50);
}

void draw() { 
  if (movingLeft || movingRight) drawThings();
}

void mousePressed() {
  if (!parsingDone) return;
  singleBarMode = !singleBarMode;
  movingLeft = singleBarMode;
  movingRight = !singleBarMode;
  stepNo = 0;
  
  if (movingLeft) {
    singleBarMode = false;
    stepNo = 50;
  }
}

void keyPressed() {  
  if (parsingDone) {
    barHue = (barHue+1) % 256;
    println("barHue = " + barHue);
    drawThings();
  }
}

void setColorForHour(int h) {  
  int steps = 4;
  if (!singleBarMode) steps *= 3;
  //int steps = singleBarMode ? 4 : 8;
  
  int idx = h*steps/24;
  int lum = 255  - idx*200/steps;
  fill(barHue, 255, lum);
}

void drawThings() {
  int cy = topMargin;
  int graphWidth = w - leftMargin - rightMargin;

  background(255);
  if (movingRight) stepNo ++;
  if (movingLeft) stepNo --;
  if (movingLeft && stepNo == 0) singleBarMode = true;

  // title
  textFont(fontBig);
  textAlign(LEFT);
  colorMode(RGB);
  fill(0,80,0);
  text("MOST POPULAR SUBJECT WORDS :: SEATTLE PUBLIC LIBRARY :: 01.02.08",
    leftMargin, topMargin - textGap
  );

  // data graph
  for (int wi = 0; wi < TOP_N; wi ++) {
    WordStat ws = popWords[wi];
    String word = ws.word;
    textAlign(RIGHT);
    fill(64);
    text(word, leftMargin - textGap, cy + textYOffset);
    
    int total = ws.getTotal();
    
    float cx = leftMargin;
    float wPerHour = graphWidth / (float)maxCount;
    colorMode(HSB);
    for (int iHour = 0; singleBarMode && iHour < 24; iHour++) {
      
      // if top word
      if (wi == 0) {
        if (iHour==12) alignNoon = cx;
        if (iHour==18) alignEvening = cx;
      } else {
        //not top word
        if (!singleBarMode) {
          if (iHour==12) cx = alignNoon;
          if (iHour==18) cx = alignEvening;
        }
      }
      
      float myW = wPerHour * ws.getFreqAtHour(iHour);
      if (myW > 0) { 
        setColorForHour(iHour);

        noStroke();
        rect(cx, cy, myW, lineHeight);
        cx += myW;
      }
    }
    colorMode(RGB);
  
    if (singleBarMode) {
      // big box
      int barW = graphWidth*total / maxCount;
      noFill();
      stroke(#000000);
      rect(leftMargin, cy, barW, lineHeight);
      textAlign(LEFT);
      fill(#008800);
      text(""+total, leftMargin + barW + 5, cy + textYOffset);    
    } else {
      //three boxes
      //noFill();
      stroke(#000000);
      float w1 = graphWidth*ws.getMorning() / maxCount;
      float w2 = graphWidth*ws.getAfternoon() / maxCount;
      float w3 = graphWidth*ws.getEvening() / maxCount;
      float f = stepNo/50f;
      f*=f;
      if (f > 1) f = 1;
      float x1 = leftMargin;
      float x2 = alignNoon*f + (leftMargin+w1)*(1-f);
      float x3 = alignEvening*f + (leftMargin+w1+w2)*(1-f);      
      colorMode(HSB);
      cx = x1;
      for (int iHour = 0; iHour < 24; iHour++) {
        if (iHour == 12) cx = x2;
        if (iHour == 18) cx = x3;
        float myW = wPerHour * ws.getFreqAtHour(iHour);
        if (myW > 0) { 
          setColorForHour(iHour);
          noStroke(); 
          rect(cx, cy, myW, lineHeight);
          cx += myW;
        }
      }
      noFill();
      stroke(#000000);
      rect(x1, cy, w1, lineHeight);
      rect(x2, cy, w2, lineHeight);
      rect(x3, cy, w3, lineHeight);
      colorMode(RGB);
      textAlign(LEFT);
      fill(#008800);
      text(""+total, x3+w3+5, cy + textYOffset);    
    }
    
    cy += lineHeight;
  } //end for each word
  
  // bottom line
  cy += 5;
  stroke(128);
  line(leftMargin, cy, w - rightMargin, cy);
  
  // x-axis markers
  int y_bottom = cy + textYOffset;
  fill(64, 64, 64, 255*stepNo/50);
  textAlign(LEFT);
  text("MORNING", leftMargin, y_bottom);    
  text("AFTERNOON", alignNoon, y_bottom);    
  text("EVENING", alignEvening, y_bottom);      
  fill(64, 64, 64, 255*(50-stepNo)/50);
  textAlign(LEFT);
  text("Frequency = 0", leftMargin, y_bottom);    
  textAlign(RIGHT);
  text("Frequency = "+maxCount, w - rightMargin, y_bottom);    
  
  if (singleBarMode) {  
    // time legend
    int boxSize = lineHeight;
    stroke(0);
    int boxX = w - rightMargin - 100; 
    int boxY = h/2;  
    colorMode(HSB);
    setColorForHour(6);
    rect(boxX, boxY, boxSize, boxSize);
    setColorForHour(12);
    rect(boxX, boxY+boxSize, boxSize, boxSize);
    setColorForHour(18);
    rect(boxX, boxY+2*boxSize, boxSize, boxSize);
  //  setColorForHour(18);
  //  rect(boxX, boxY+3*boxSize, boxSize, boxSize);
    colorMode(RGB);
    fill(64);
    textAlign(RIGHT);
    int textX = w - rightMargin;
    text("6AM-12PM", textX, boxY + textYOffset);
    text("12PM-6PM", textX, boxY + textYOffset + boxSize);
    text("6PM-12AM", textX, boxY + textYOffset + boxSize*2);
  //  text("18:00-24:00", textX, boxY + textYOffset + boxSize*3);
  }

  //movingLeft = movingRight = stepNo != 50;
  if (stepNo >= 50) movingRight = false;
  if (stepNo == 0) movingLeft = false; 
}  

boolean blackListed(String word) {
  return
    word.equals("a") ||
    word.equals("an") ||
    word.equals("the") ||
    word.equals("for") ||
    word.equals("and") ||
    word.equals("or") ||
    word.equals("etc") ||
    word.equals("a") ||
    false; 
}

void recordThisWord(String word, int hour) {
  word = word.toLowerCase();
  
  if (blackListed(word)) return;
  
  if (wordMap == null) wordMap = new HashMap();
  WordStat ws = (WordStat) wordMap.get(word);
  if (ws==null) {
    ws = new WordStat(word);
    wordMap.put(word, ws);
  }
  ws.addFreqAtHour(hour);
  
  if (count <= 20) {
    println("  " + ws);
  }
}

void parseTransaction(XMLElement eTran, int hour) {
  XMLElement ev2 = eTran.getChild(1); 
  XMLElement[] eSubjects = ev2.getChildren();
  if (eSubjects == null) return;
  for (int i = 0; i < eSubjects.length; i++) {
    XMLElement eSubject = eSubjects[i];
    String sentence = eSubject.firstChild().toString();
    count ++;
    if (count <= 20) { 
      println(sentence);
    }

    String[] words = split(sentence, " ");
    for (int iw = 0; iw < words.length; iw++) {
      wordCount ++;
      String thisWord = words[iw];
      recordThisWord(thisWord, hour);
    }            
    
  }
}

void xmlEvent(XMLElement element) {  
  background(#FFFFFF);
  println("root count: " + element.countChildren());
  XMLElement eDay = element.firstChild();
  println("day count: " + eDay.countChildren());
  
  XMLElement[] eHours = eDay.getChildren();
  println("eHours.length = " + eHours.length);
  for (int iHour = 0; iHour < eHours.length; iHour++) {
    XMLElement[] eTrans = eHours[iHour].getChildren();
    println("Hour #"+iHour+" eTrans.length=" + eTrans.length); 
    for (int iTran = 0;  iTran < eTrans.length; iTran++) {
      parseTransaction(eTrans[iTran], iHour);
    }
  }
  
  println("");
  println("total words = " + wordCount);
  println("total distinct words = " + wordMap.size());
  calculateStatistics();
  parsingDone = true;
  drawThings();
}

void xmlEvent_old(XMLElement element) {  // this method is called when the main (or root) xml element is finished loading. 'element' represents the root xml element
  XMLElement[] transactions;         // create an array of XMLElements named 'e'
  XMLElement[] aTransaction;
  int i,j;
  
  background(#FFFFFF);

  count = 0;
  wordCount = 0;
  transactions = element.getChildren();                                        // get all the child elements of the root element and store them in the array we just created  
  for(i=0; i<transactions.length; i++) {                                               // loop for as long as i is less than the length of the transactions array, with i starting at 0
    aTransaction = transactions[i].getChildren();                              // get the children of the current transaction
    int min = 0;
    for(j=0; j<aTransaction.length; j++) {        // loop through all the children of the current transaction
      XMLElement subjects = aTransaction[j];
      if (subjects.getName().equals("ckotime")) {
        String timeString = "" + subjects.firstChild();
        String[] tokens = split(timeString, ':');
        String mins = tokens[1];
        min = int(mins);
      }
      if (subjects.getName().equals("subjects")) {
        XMLElement[] allSubjects = subjects.getChildren();
        for (int k = 0; k < allSubjects.length; k++) { 
          XMLElement subject = allSubjects[k];
          if (subject.hasChildren()) {
            String sentence = subject.firstChild().toString();
            count ++;
            if (count <= 20) { 
              println(sentence);
            }

            String[] words = split(sentence, " ");
            for (int iw = 0; iw < words.length; iw++) {
              wordCount ++;
              String thisWord = words[iw];
              recordThisWord(thisWord, min);
            }            
            
          }
        }
      }
    }
  }
  println("");
  println("total words = " + wordCount);
  println("total distinct words = " + wordMap.size());
  calculateStatistics();
  parsingDone = true;
  drawThings();
}

void calculateStatistics() {
  Object[] o = wordMap.values().toArray();
  WordStat[] a = new WordStat[o.length];
  for (int i = 0; i < o.length; i++) { 
    a[i] = (WordStat) (o[i]);
  }
  Arrays.sort(a);
  reverse(a);
  println("Most popular words:");
  for (int i = 0; i < 10; i ++) {
    println("  " + a[i]);
  }
  maxCount = a[0].getTotal();
  popWords = a;
}

void reverse(Object[] a) {
  int len = a.length;
  int len2 = len/2;
  for (int i = 0; i < len2; i++) {
    int k = len - i - 1;
    Object tmp = a[i];
    a[i] = a[k];
    a[k] = tmp;
  }
}

class WordStat implements Comparable {
  String word; // this should be private, but whatever ...
  private int[] hours;
  private int total;
  
  public WordStat(String word) {
    this.word = word;
    hours = new int[24];
    for (int i = 0; i < 24; i++) hours[i] = 0;
  }
  
  public void addFreqAtHour(int hour) {
    hours[hour]++;
    total ++;
  }
    
  public int getTotal() {
    return total;
  }
  
  public int getFreqAtHour(int i) {
    return hours[i];
  }
  
  public int getMorning() {
    int r = 0;
    for (int i = 6; i < 12; i++) r+=hours[i];
    return r;
  }
  
  public int getAfternoon() {
    int r = 0;
    for (int i = 12; i < 18; i++) r+=hours[i];
    return r;
  }
  
  public int getEvening() {
    int r = 0;
    for (int i = 18; i < 24; i++) r+=hours[i];
    return r;
  }
  
  public int compareTo(Object o) {
    WordStat other = (WordStat) o; 
    return total - other.total;
  }
  
  public String toString() {
    float percent = getTotal()*100/(float)wordCount;
    float pci = round(percent);
    return word + "(" + getTotal() + " " + pci + "%)";
  }
}
